// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#define _USE_MATH_DEFINES
#else
#include <pwd.h>
#endif

#include <math.h>
#include <time.h>
#include <dirent.h>
#include <sys/types.h>
#include <unistd.h>
#include <limits.h>
#include <stdio.h>
#include <sys/stat.h>

#include "util.h"

#include "sim.h"
#include "dos.h"
#include "server.h"
#include "elec_plugins.h"
#include "light_interface.h"
#include "dump.h"
#include "complex_solver.h"
#include "elec_plugins.h"
#include "license.h"
#include "inp.h"
#include "gui_hooks.h"
#include "timer.h"
#include "rand.h"
#include "hard_limit.h"
#include "patch.h"

static int unused __attribute__ ((unused));

struct device cell;

void device_init(struct device *in)
{
	in->Voc = 0.0;
	in->Jsc = 0.0;
	in->FF = 0.0;
	in->Pmax = 0.0;
	in->Pmax_voltage = 0.0;
	device_get_memory(in);

}

int main(int argc, char *argv[])
{
	timer_init(0);
	timer_init(1);
	dbus_init();
	set_ewe_lock_file("", "");
	cell.onlypos = FALSE;

	char pwd[1000];
	getcwd(pwd, 1000);

	remove("snapshots.zip");
	remove("light_dump.zip");

	hard_limit_init();

	set_plot_script_dir(pwd);

	if (scanarg(argv, argc, "--help") == TRUE) {
		printf("opvdm_core - Organic Photovoltaic Device Model\n");
		printf(copyright);
		printf("\n");
		printf("Usage: opvdm_core [options]\n");
		printf("\n");
		printf("Options:\n");
		printf("\n");
		printf("\t--outputpath\toutput data path");
		printf("\t--inputpath\t sets the input path\n");
		printf("\t--version\tdisplays the current version\n");
		printf("\t--zip_results\t zip the results\n");
		printf("\t--optics\t runs only optical simulation\n");
		printf("\t--cpus\t sets the number of CPUs\n");
		printf("\n");
		printf
		    ("Additional information about opvdm is available at www.opvdm.com.\n");
		printf("\n");
		printf
		    ("Report bugs to: roderick.mackenzie@nottingham.ac.uk\n\n");
		exit(0);
	}
	if (scanarg(argv, argc, "--version") == TRUE) {
		printf("opvdm_core, Version %s\n", opvdm_ver);
		printf(copyright);
		printf(this_is_free_software);
		printf
		    ("There is ABSOLUTELY NO WARRANTY; not even for MERCHANTABILITY or\n");
		printf("FITNESS FOR A PARTICULAR PURPOSE.\n");
		printf("\n");
		exit(0);
	}

#ifndef windows
	if (geteuid() == 0) {
		ewe("Don't run me as root!\n");
	}
#endif

	set_dump_status(dump_stop_plot, FALSE);
	set_dump_status(dump_print_text, TRUE);

	set_io_dump(FALSE);
	srand(time(0));
	textcolor(fg_green);
	randomprint("Organic Photovoltaic Device Model (www.opvdm.com)\n");
	randomprint(copyright);
#ifdef gnu_copyright
	randomprint
	    ("You should have received a copy of the GNU General Public License\n");
	randomprint
	    ("along with this software.  If not, see www.gnu.org/licenses/.\n");
#endif
	randomprint("\n");
	randomprint
	    ("If you wish to collaborate in anyway please get in touch:\n");
	randomprint("roderick.mackenzie@nottingham.ac.uk\n");
	randomprint("www.roderickmackenzie.eu/contact.html\n");
	randomprint("\n");
	textcolor(fg_reset);

	globalserver.on = FALSE;
	globalserver.cpus = 1;
	globalserver.readconfig = TRUE;

	if (scanarg(argv, argc, "--outputpath") == TRUE) {
		strcpy(sim_output_path(),
		       get_arg_plusone(argv, argc, "--outputpath"));
	} else {
		strcpy(sim_output_path(), pwd);
	}

	if (scanarg(argv, argc, "--inputpath") == TRUE) {
		strcpy(sim_input_path(),
		       get_arg_plusone(argv, argc, "--inputpath"));
	} else {
		strcpy(sim_input_path(), sim_output_path());
	}

	dump_init(&cell);
	dump_load_config(&cell);

	if (scanarg(argv, argc, "--onlypos") == TRUE) {
		cell.onlypos = TRUE;
	}

	char name[200];
	struct inp_file inp;
	inp_init(&inp);
	inp_load_from_path(&inp, sim_input_path(), "ver.inp");
	inp_check(&inp, 1.0);
	inp_search_string(&inp, name, "#core");
	inp_free(&inp);

	if (strcmp(name, opvdm_ver) != 0) {
		printf
		    ("Software is version %s and the input files are version %s\n",
		     opvdm_ver, name);
		exit(0);
	}

	gui_start();
	if (scanarg(argv, argc, "--optics") == FALSE)
		server_init(&globalserver);

	if (scanarg(argv, argc, "--lock") == TRUE) {
		server_set_dbus_finish_signal(&globalserver,
					      get_arg_plusone(argv, argc,
							      "--lock"));
	}

	int ret = 0;

	int do_fit = FALSE;

	if (scanarg(argv, argc, "--fit") == TRUE)
		do_fit = TRUE;

	FILE *f = fopen("fit.inp", "r");
	if (f != NULL) {
		fclose(f);
		inp_init(&inp);
		inp_load_from_path(&inp, pwd, "fit.inp");
		int fit_temp;
		inp_search_int(&inp, &fit_temp, "#do_fit");
		if (fit_temp == 1) {
			do_fit = TRUE;
		}

		inp_free(&inp);
	}

	if (do_fit == TRUE) {
		set_dump_status(dump_lock, FALSE);
		set_dump_status(dump_print_text, FALSE);
		set_dump_status(dump_iodump, FALSE);
		set_dump_status(dump_optics, FALSE);
		set_dump_status(dump_newton, FALSE);
		set_dump_status(dump_plot, FALSE);
		set_dump_status(dump_stop_plot, FALSE);
		set_dump_status(dump_opt_for_fit, FALSE);
		set_dump_status(dump_print_newtonerror, FALSE);
		set_dump_status(dump_print_converge, FALSE);
		set_dump_status(dump_print_pos_error, FALSE);
		set_dump_status(dump_lock, TRUE);
	}
#include "main_args.c"
	if (scanarg(argv, argc, "--optics") == TRUE) {
		gui_start();
		struct light two;
		light_init(&two, &cell, pwd);

		light_load_config(&two);
		two.disable_transfer_to_electrical_mesh = TRUE;
		set_dump_status(dump_lock, FALSE);
		set_dump_status(dump_optics, TRUE);
		set_dump_status(dump_optics_verbose, TRUE);
		double Psun;
		inp_init(&inp);
		inp_load_from_path(&inp, pwd, "light.inp");
		inp_search_double(&inp, &(Psun), "#Psun");
		Psun = 1.0;
		inp_free(&inp);

		light_solve_and_update(&cell, &two, Psun, 0.0);
		light_dump(&two);
		light_free(&two);
		complex_solver_free();
	} else {
		gen_dos_fd_gaus_fd();

		server_add_job(&globalserver, cell.outputpath, cell.inputpath);
		print_jobs(&globalserver);
		ret = server_run_jobs(&globalserver);

	}

	server_shut_down(&globalserver);

	if (scanarg(argv, argc, "--zip_results") == TRUE) {
		printf("zipping results\n");
		int ret;
		char temp[200];
		DIR *dir = opendir("snapshots");
		if (dir) {
			closedir(dir);
			ret =
			    system("zip -r -j -q snapshots.zip ./snapshots/*");
			if (ret == -1) {
				printf("tar returned error\n");
			}

			join_path(2, temp, cell.outputpath, "snapshots");
			remove_dir(temp);

		}

		dir = opendir("light_dump");
		if (dir) {
			closedir(dir);
			ret =
			    system
			    ("zip -r -j -q light_dump.zip ./light_dump/*");
			if (ret == -1) {
				printf("tar returned error\n");
			}

			join_path(2, temp, cell.outputpath, "light_dump");
			remove_dir(temp);

		}

	}

	hard_limit_free();
	if (ret != 0) {
		return 1;
	}
	return 0;
}
