// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "sim.h"
#include "dump.h"
#include "ntricks.h"
#include "../../gui_hooks.h"

static int unused __attribute__ ((unused));

static double glob_wanted = 0.0;

struct newton_math_state math_save_state;

void newton_push_state(struct device *in)
{
	math_save_state.min_cur_error = in->min_cur_error;
	math_save_state.max_electrical_itt = in->max_electrical_itt;
	math_save_state.newton_min_itt = in->newton_min_itt;
	math_save_state.electrical_clamp = in->electrical_clamp;
	math_save_state.newton_clever_exit = in->newton_clever_exit;
}

void newton_pop_state(struct device *in)
{
	in->min_cur_error = math_save_state.min_cur_error;
	in->max_electrical_itt = math_save_state.max_electrical_itt;
	in->newton_min_itt = math_save_state.newton_min_itt;
	in->electrical_clamp = math_save_state.electrical_clamp;
	in->newton_clever_exit = math_save_state.newton_clever_exit;
}

void ramp_externalv(struct device *in, double from, double to)
{
	double V = from;
	double dV = 0.1;
	if ((to - from) < 0.0)
		dV *= -1.0;
	printf("dV=%le\n", dV);
	printf("Ramping: from=%e to=%e\n", from, to);

	if (fabs(to - from) <= fabs(dV))
		return;

	do {
		V += dV;
		if (get_dump_status(dump_print_text) == TRUE)
			printf("ramp: %lf %lf %d\n", V, to, in->kl_in_newton);
		sim_externalv(in, V);

		plot_now(in, "jv.plot");
		gui_send_data("pulse");

		if (fabs(in->Vapplied - to) < fabs(dV)) {
			break;
		}

	} while (1);

	if (V != to) {
		V = to;
		sim_externalv(in, V);
	}

	return;
}

void ramp(struct device *in, double from, double to, double steps)
{
	in->kl_in_newton = FALSE;
	solver_realloc(in);

	in->Vapplied = from;
	newton_push_state(in);
	double dV = 0.20;
	in->min_cur_error = 1e-5;
	in->max_electrical_itt = 12;
	in->newton_min_itt = 3;
	in->electrical_clamp = 2.0;
	in->newton_clever_exit = FALSE;
	if ((to - from) < 0.0)
		dV *= -1.0;
	printf("dV=%le\n", dV);
	printf("Ramping: from=%e to=%e\n", from, to);

	if (fabs(to - from) <= fabs(dV))
		return;

	do {
		in->Vapplied += dV;

		if (get_dump_status(dump_print_text) == TRUE)
			printf("ramp: %lf %lf %d\n", in->Vapplied, to,
			       in->kl_in_newton);
		solve_all(in);
		plot_now(in, "jv_vars.plot");

		if (fabs(in->Vapplied - to) < fabs(dV)) {

			break;
		}

	} while (1);

	newton_pop_state(in);

	if (in->Vapplied != to) {
		in->Vapplied = to;
		solve_all(in);
	}

	printf("Finished with ramp\n");
	return;
}

void save_state(struct device *in, double to)
{

	printf("Dumping state\n");
	int i;
	int band;
	FILE *state;
	state = fopena(in->outputpath, "state.dat", "w");

	fprintf(state, "%le ", to);

	for (i = 0; i < in->ymeshpoints; i++) {
		fprintf(state, "%le %le %le ", in->phi[i], in->x[i], in->xp[i]);

		for (band = 0; band < in->srh_bands; band++) {
			fprintf(state, "%le %le ", in->xt[i][band],
				in->xpt[i][band]);
		}

	}
	fclose(state);

}

int load_state(struct device *in, double voltage)
{

	printf("Load state\n");
	int i;
	int band;
	double vtest;
	FILE *state;
	state = fopena(in->outputpath, "state.dat", "r");
	if (!state) {
		printf("State not found\n");
		return FALSE;
	}

	unused = fscanf(state, "%le", &vtest);
	printf("%le %le", voltage, vtest);
	if (vtest != voltage) {
		printf("State not found\n");
		return FALSE;
	}
	printf("Loading state\n");

	in->Vapplied = vtest;

	for (i = 0; i < in->ymeshpoints; i++) {
		unused =
		    fscanf(state, "%le %le %le ", &(in->phi[i]), &(in->x[i]),
			   &(in->xp[i]));

		for (band = 0; band < in->srh_bands; band++) {
			unused =
			    fscanf(state, "%le %le ", &(in->xt[i][band]),
				   &(in->xpt[i][band]));
		}

	}
	fclose(state);
	return TRUE;

}

void newton_aux_i(struct device *in, double V, double *i, double *didv,
		  double *didphi, double *didxil, double *didxipl,
		  double *didphir, double *didxir, double *didxipr)
{

	double i0 = *i;
	double didv0 = *didv;
	double didphi0 = *didphi;
	double didxil0 = *didxil;
	double didxipl0 = *didxipl;
	double didphir0 = *didphir;
	double didxir0 = *didxir;
	double didxipr0 = *didxipr;

	*i = glob_wanted - i0 - V / in->Rshunt;
	*didv = -didv0 - 1.0 / in->Rshunt;
	*didphi = -didphi0;
	*didxil = -didxil0;
	*didxipl = -didxipl0;
	*didphir = -didphir0;
	*didxir = -didxir0;
	*didxipr = -didxipr0;

	return;
}

double sim_i(struct device *in, double wantedi)
{

	in->kl_in_newton = TRUE;
	solver_realloc(in);
	glob_wanted = wantedi;
	in->newton_aux = &newton_aux_i;
	solve_all(in);

	return 0.0;
}

void newton_aux_externalv(struct device *in, double V, double *i, double *didv,
			  double *didphi, double *didxil, double *didxipl,
			  double *didphir, double *didxir, double *didxipr)
{
	double i0 = *i;
	double didv0 = *didv;
	double didphi0 = *didphi;
	double didxil0 = *didxil;
	double didxipl0 = *didxipl;
	double didphir0 = *didphir;
	double didxir0 = *didxir;
	double didxipr0 = *didxipr;

	*i = glob_wanted - (V + in->Rcontact * (i0 + V / in->Rshunt));
	*didv = -(1.0 + in->Rcontact * (didv0 + 1.0 / in->Rshunt));
	*didphi = -(in->Rcontact * (didphi0));
	*didxil = -(in->Rcontact * (didxil0));
	*didxipl = -(in->Rcontact * (didxipl0));
	*didphir = -(in->Rcontact * (didphir0));
	*didxir = -(in->Rcontact * (didxir0));
	*didxipr = -(in->Rcontact * (didxipr0));
	return;
}

double sim_externalv_ittr(struct device *in, double wantedv)
{
	double clamp = 0.1;
	double step = 0.01;
	double e0;
	double e1;
	double i0;
	double i1;
	double deriv;
	double Rs = in->Rcontact;
	solve_all(in);
	i0 = get_I(in);

	double itot = i0 + in->Vapplied / in->Rshunt;

	e0 = fabs(itot * Rs + in->Vapplied - wantedv);
	in->Vapplied += step;
	solve_all(in);

	i1 = get_I(in);
	itot = i1 + in->Vapplied / in->Rshunt;

	e1 = fabs(itot * Rs + in->Vapplied - wantedv);

	deriv = (e1 - e0) / step;
	step = -e1 / deriv;

	in->Vapplied += step;
	int count = 0;
	int max = 1000;
	do {
		e0 = e1;
		solve_all(in);
		itot = i1 + in->Vapplied / in->Rshunt;
		e1 = fabs(itot * Rs + in->Vapplied - wantedv);

		deriv = (e1 - e0) / step;
		step = -e1 / deriv;

		step = step / (1.0 + fabs(step / clamp));
		in->Vapplied += step;
		if (count > max)
			break;
		count++;
	} while (e1 > 1e-8);

	double ret = get_I(in) + in->Vapplied / in->Rshunt;

	return ret;
}

double sim_externalv(struct device *in, double wantedv)
{

	in->kl_in_newton = TRUE;
	solver_realloc(in);

	glob_wanted = wantedv;

	in->newton_aux = &newton_aux_externalv;

	solve_all(in);

	return 0.0;
}

void solve_all(struct device *in)
{
	solve_cur(in);
}

void newton_sim_jv(struct device *in)
{
	in->kl_in_newton = FALSE;
	solver_realloc(in);

	solve_all(in);
}
