// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "sim.h"
#include "inp.h"

static int unused __attribute__ ((unused));

static double *sun;
static double *voltage;
static double *laser;
static double *time_mesh;
static double *fs_laser;
static int mesh_len = 0;
static int use_mesh = FALSE;
static int mesh_pos = 0;

void time_load_mesh(struct device *in)
{
	int i;
	mesh_pos = 0;
	struct inp_file inp;
	inp_init(&inp);
	inp_load_from_path(&inp, in->inputpath, "time_mesh.inp");
	inp_check(&inp, 1.0);

	inp_reset_read(&inp);

	sscanf(inp_get_string(&inp), "%d", &mesh_len);

	time_mesh = (double *)malloc(mesh_len * sizeof(double));
	sun = (double *)malloc(mesh_len * sizeof(double));
	voltage = (double *)malloc(mesh_len * sizeof(double));
	laser = (double *)malloc(mesh_len * sizeof(double));
	fs_laser = (double *)malloc(mesh_len * sizeof(double));

	for (i = 0; i < mesh_len; i++) {
		sscanf(inp_get_string(&inp), "%le %le %le %le %le",
		       &time_mesh[i], &laser[i], &sun[i], &voltage[i],
		       &fs_laser[i]);
	}

	in->time = time_mesh[0];
	in->dt = time_mesh[1] - time_mesh[0];

	inp_free(&inp);

	use_mesh = TRUE;
}

void time_init(struct device *in)
{
	int i;
	int band;

	for (i = 0; i < in->ymeshpoints; i++) {

		for (band = 0; band < in->srh_bands; band++) {
			in->ntlast[i][band] = in->nt[i][band];
			in->ptlast[i][band] = in->pt[i][band];
		}

		in->nlast[i] = in->n[i];
		in->plast[i] = in->p[i];

	}
}

void device_timestep(struct device *in)
{
	int i;
	int band;

	for (i = 0; i < in->ymeshpoints; i++) {
		in->nlast[i] = in->n[i];
		in->plast[i] = in->p[i];

		for (band = 0; band < in->srh_bands; band++) {
			in->ntlast[i][band] = in->nt[i][band];
			in->ptlast[i][band] = in->pt[i][band];
		}
	}
	in->Vapplied_last = in->Vapplied;
	in->VCext_last = in->VCext;
	in->Ilast = get_I(in);

	if (use_mesh == TRUE) {
		if (mesh_pos < (mesh_len - 1)) {
			in->time = time_mesh[mesh_pos];
			in->dt =
			    (time_mesh[mesh_pos + 1] - time_mesh[mesh_pos]);
			mesh_pos++;
		}
	} else {
		in->time += in->dt;
	}

}

int time_run()
{
	if (mesh_pos < (mesh_len - 1)) {
		return TRUE;
	} else {
		return FALSE;
	}
}

double time_get_voltage()
{
	return voltage[mesh_pos];
}

double time_get_fs_laser()
{
	return fs_laser[mesh_pos];
}

double time_get_sun()
{
	return sun[mesh_pos];
}

double time_get_laser()
{
	return laser[mesh_pos];

}

void time_memory_free()
{
	free(time_mesh);
	free(sun);
	free(voltage);
	free(laser);
	free(fs_laser);
}
